# -*- coding: utf-8 -*-
from datetime import date, timedelta
import os, time, decimal

from django.conf import settings as django_settings
from django.db import models, connection, transaction
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ugettext as ugt
from django.contrib.auth.models import User
from django.core.files.storage import FileSystemStorage
from django.core.mail import send_mail
from django.db.models import Q
from django.core.exceptions import PermissionDenied

from jumbo.db import models as jumbo_models
from jumbo.middleware.thread_local import get_current_user
from jumbo.utils.functions import get_settings
from jumbo.db.models import Settings as JSettings
from jumbo.db.models import DateModel, UserModel, StatusModel, OrderedModel
from jumbo.apps.project.models import Project
from jumbo.apps.i18n.models import I18NModel, TranslationModel

from organization.models import Organization, OrganizationAddress, Contact
from ticket import settings
from ticket.email import TicketCloseMail
#DROP TABLE ticket_category CASCADE;
#DROP TABLE ticket_categorytranslation CASCADE;
#DROP TABLE ticket_log CASCADE;
#DROP TABLE ticket_settings CASCADE;
#DROP TABLE ticket_ticket CASCADE;
#DROP TABLE ticket_ticket_assigned_to CASCADE;
#DROP TABLE ticket_ticketassistance CASCADE;
#DROP TABLE ticket_ticketdocuments CASCADE;
#DROP TABLE ticket_ticketproject CASCADE;
#DROP TABLE ticket_ticketproject_managers CASCADE;
#DROP TABLE ticket_ticketproject_staff CASCADE;
#DROP TABLE ticket_ticketuser CASCADE;
#DROP TABLE ticket_tipology CASCADE;
#DROP TABLE ticket_tipologytranslation CASCADE;

STATUS_CODES = (
    (0, _('closed')),
    (1, _('open')),
    (2, _('standby')),
    (3, _('waiting for input')),
    (4, _('not repeatable')),
    (5, _('won\'t fix')),
    (6, _('working')),
    (7, _('assigned')),
    (8, _('to discuss')),
)

PRIORITY_CODES = (
    (0, _('very urgent')),
    (1, _('urgent')),
    (2, _('important')),
    (3, _('normal')),
    (4, _('low')),
    (5, _('someday')),
)

###############################
# TICKET USER                 #
###############################
#class TicketUserManager(models.Manager):
#    def get_query_set(self):
#        qs = super(TicketUserManager, self).get_query_set()
#
#        user = get_current_user()
#
#        if not user:
#            return qs.none()
#
#        if not user.is_authenticated():
#            return qs.none()
#
#        #if user.is_superuser or user.has_perm('ticket.list_all_ticketusers'):
#        #    return qs
#
#        return qs.filter(user=user)


class TicketUser(models.Model):
    user = models.OneToOneField (
        User,
        db_index = True,
        related_name = 'ticket_user',
        verbose_name = _("User")
    )
    enabled = models.BooleanField (
        db_index = True,
        verbose_name = _("Enabled")
    )
    is_guest = models.BooleanField (
        db_index = True,
        verbose_name = _("Is Guest"),
        help_text = _("If True you won't see the User in the 'Assigned To' Select.")
    )
    #objects = TicketUserManager()

    class Meta:
        verbose_name = _("Ticket User")
        verbose_name_plural = _("Ticket Users")
        ordering = ('user__last_name', 'user__first_name')
        permissions = (
            ("view_ticketuser",        "Can view ticketuser"),
            ("list_all_ticketusers",   "Can list all ticketusers"),
        )

    def small_name(self):
        if self.user.last_name or self.user.first_name:
            return u"%s %s." % (self.user.last_name, self.user.first_name[:1].upper())
        return self.user.username

    def __unicode__(self):
        if self.user.last_name or self.user.first_name:
            return u"%s %s" % (self.user.last_name, self.user.first_name)
        return self.user.username


############################
# TICKET                   #
############################
class AllTicketManager(models.Manager):
    def get_query_set(self):
        qs = super(AllTicketManager, self).get_query_set()
        return qs

class TicketManager(models.Manager):
    def get_query_set(self):
        qs = super(TicketManager, self).get_query_set()
        user = get_current_user()

        if not user:
            return qs.none()

        if not user.is_authenticated():
            return qs.none()

        if user.is_superuser or user.has_perm('ticket.list_all_tickets'):
            return qs

        try:
            ticket_user = user.ticket_user
        except TicketUser.DoesNotExist:
            return qs.none()

        #if user.has_perm("ticket.list_own_organization_tickets"):
        # Se l'utente è un ticket user guest gli faccio vedere solo
        # i ticket associati alla sua anagrafica
        #
        if ticket_user.enabled and ticket_user.is_guest:
            try:
                if user.contact and user.contact.organization:
                    return qs.filter(public=True, organization = user.contact.organization)
            except Contact.DoesNotExist:
                return qs.none()

        id_projects = [ p.id for p in TicketProject.objects.all() ]

        q1 = Q(assigned_to=ticket_user) | Q(submitter=user)
        if user.has_perm("ticket.view_ticket_not_assigned"):
            q2 = Q(assigned_to__isnull=True)
            q3 = Q(project__managers=ticket_user)
            qs = qs.filter(q1 | q2 | q3)
        else:
            q2 = Q(assigned_to__isnull=True) & Q(project__in=id_projects)
            q3 = Q(project__managers=ticket_user)
            qs = qs.filter(q1|q2|q3)

        qs = qs.distinct()
        return qs


class Ticket(DateModel, UserModel):
    ## ForeignKey
    project = models.ForeignKey (
        'TicketProject',
        null = True, blank = True,
        db_index = True,
        related_name = "tickets",
        verbose_name = _("Project"),
    )
    organization = models.ForeignKey (
        Organization,
        null = True, blank = True,
        db_index = True,
        related_name = "tickets",
        verbose_name = _("Organization"),
    )
    address = models.ForeignKey (
        OrganizationAddress,
        related_name = 'tickets',
        null = True, blank = True,
        db_index = True,
        verbose_name = _("Address"),
    )
    contact = models.ForeignKey (
        Contact,
        related_name = 'tickets',
        null = True, blank = True,
        db_index = True,
        verbose_name = _("Contact"),
    )
    submitter = models.ForeignKey (
        User,
        db_index = True,
        related_name = "tickets_summitted",
        verbose_name = _("Submitter"),
    )
    assigned_by = models.ForeignKey (
        TicketUser,
        blank = True, null = True,
        db_index = True,
        related_name = "tickets_assigned_by_me",
        verbose_name = _("Assigned by"),
    )
    assigned_to = models.ManyToManyField (
        TicketUser,
        blank = True, null = True,
        db_index = True,
        related_name = "tickets_assigned_to_me",
        verbose_name = _("Assigned to"),
    )
    category = models.ForeignKey (
        "Category",
        db_index = True,
        blank = True, null = True,
        related_name = "tickets",
        verbose_name = _("Category"),
    )
    tipology = models.ForeignKey (
        "Tipology",
        db_index = True,
        default = 1,
        related_name = "tickets",
        verbose_name = _("Tipology"),
    )
    # Campi
    status = models.IntegerField (
        default = 1,
        db_index = True,
        choices = STATUS_CODES,
        verbose_name = _("Status"),
    )
    title = models.CharField (
        max_length = 100,
        verbose_name = _("Title"),
    )
    keywords = models.CharField (
        max_length=150,
        verbose_name = _("Keywords"),
        help_text="E' necessario mettere alcune parole utili " +\
             "per la catalogazione del ticket"
    )
    description = jumbo_models.HTMLField (
        verbose_name = _("Description"),
    )
    priority = models.IntegerField (
        default = 1,
        db_index = True,
        choices = PRIORITY_CODES,
        verbose_name = _("Priority"),
    )
    duration  = models.IntegerField (
        null = True, blank = True,
        verbose_name = _("Duration"),
        help_text = _("In Minutes")
    )
    deadline = models.DateField(
        null = True, blank = True,
        verbose_name = _("Deadline"),
    )
    note = models.TextField (
        verbose_name = _("Note"),
        null = True, blank = True,
    )
    public = models.BooleanField(
        verbose_name = _("Public"),
        help_text = _("If ticket can be view by guest"),
        default = False,
    )

    objects = TicketManager()
    all_tickets = AllTicketManager()

    def get_from_mail(self):
        email = get_settings("ticket.settings", "GENERIC_MAIL_TICKET")
        if self.project and self.project.assistance_email:
            email = self.project.assistance_email
        return email

    def get_email_assigned_to(self):
        if self.assigned_to:
            return [ u.user.email for u in self.assigned_to.all() if u.user.email ]
        else:
            if self.project:
                return [ u.user.email for u in self.project.get_list_staff() if u.user.email ]

    def get_string_assigned_to(self):
        l_assignedto = []
        u = get_current_user()
        for a in self.assigned_to.all():
            if a == u:
                l_assignedto.insert(0, a.user.get_full_name() or a.user.username)
            else:
                l_assignedto.append(a.user.get_full_name() or a.user.username)
        if l_assignedto:
            return ', '.join(l_assignedto)
        return ''
#        if self.assigned_to:
#            assigned_to = [ m.get_full_name() or m.username for m in self.assigned_to.all() ]
#            return ', '.join(assigned_to)
#        return ''
    string_assigned_to = property(get_string_assigned_to)
    get_string_assigned_to.short_description = _('Assigned To')


    def get_all_address_to_send_email(self):
        s = set()
        if self.submitter.email:
            s.add(self.submitter.email)

        for email in self.get_email_assigned_to():
            s.add(email)

        if self.project:
            for email in self.project.get_email_managers():
                s.add(email)
        return s

    def is_closed(self):
        if self.status in (0, 5):
            return True
        return False


    def get_total_assistances(self):
        return self.assistances.count()
    total_assistances=property(get_total_assistances)
    get_total_assistances.short_description = _("Total Assistances")


    def get_comments_count(self):
        from django.contrib.contenttypes.models import ContentType
        from django.contrib.comments.models import Comment
        ctype = ContentType.objects.get_for_model(self)
        num_comments = Comment.objects.filter(content_type=ctype, is_public = True, object_pk=str(self.id)).count()
        return num_comments


    def __unicode__(self):
        return unicode(self.title)


    def save(self, force_insert=False, force_update=False):
        from jumbo.middleware.thread_local import get_current_user
        user = get_current_user()
        if not self.id:
            self.submitter = user
            # Se non c'e il progetto e' perche il ticket e' stato inserito
            # dall'utente guest. In questo caso vado a cercare il progetto
            # di default per quella anagrafica, altrimenti prendo il primo
            # progetto disponibile sempre di quell'anagrafica
            if not self.project:
                project = None
                org_project = TicketProject.objects.filter(organization=self.organization, default_for_organization=True)
                if org_project:
                    project = org_project[0]
                else:
                    org_project = TicketProject.objects.filter(organization=self.organization)
                    if org_project:
                        project = org_project[0]
                self.project = project
        super(Ticket, self).save(force_insert=force_insert, force_update=force_update)


    def send_close_mail(self):
        send_mail = get_settings("ticket.settings", "SEND_MAIL")
        if send_mail:
            TicketCloseMail(self).send()


    class Meta:
        verbose_name = _("Ticket")
        verbose_name_plural = _("Tickets")
        permissions = (
            ("view_ticket"                   , "Can view tickets"),
            ("view_ticket_not_assigned"      , "Can view tickets not assigned"),
            ("can_reopen_tickets"            , "Can reopen tickets"),
            ("list_all_tickets"              , "Can list all tickets"),
            ("list_own_organization_tickets" , "Can list own organization tickets"),
        )


class TicketDocuments(DateModel, UserModel, StatusModel):
    ## ForeignKey
    ticket = models.ForeignKey (
        Ticket,
        db_index = True,
        related_name = "documents",
        verbose_name = _("Ticket"),
    )
    file = models.FileField (
        upload_to = settings.TICKET_PATH_DOCUMENTS,
        verbose_name = _("File"),
    )
    description = models.CharField(
        max_length = 255,
        verbose_name = _("Description"),
        blank = True, null = True
    )

############################
# PROGETTO                 #
############################
PSTATUS_CODES = (
    (1, _('Open')),
    (0, _('Closed')),
)

class AllProjectManager(models.Manager):
    def get_query_set(self):
        return super(AllProjectManager, self).get_query_set()

class ProjectManager(models.Manager):
    def get_query_set(self):
        qs = super(ProjectManager, self).get_query_set()

        user = get_current_user()

        if not user:
            return qs.none()

        if not user.is_authenticated():
            return qs.none()

        if user.is_superuser or user.has_perm('ticket.list_all_projects'):
            return qs

        try:
            ticket_user = user.ticket_user
        except TicketUser.DoesNotExist:
            return qs.none()

        #if user.has_perm("ticket.list_own_organization_tickets"):
        # Se l'utente è un ticket user guest gli faccio vedere solo
        # i ticket associati alla sua anagrafica
        #
        if ticket_user.enabled and ticket_user.is_guest:
            try:
                if user.contact and user.contact.organization:
                    return qs.filter(organization = user.contact.organization)
            except Contact.DoesNotExist:
                return qs.none()

        try:
            qs = qs.filter(Q(managers=user.ticket_user) | Q(staff=user.ticket_user))
        except:
            return qs.none()

        qs = qs.distinct()
        return qs


class TicketProject(Project):
    """Project: each user can have multiple projects"""
    ## ForeignKey
    managers = models.ManyToManyField (
        TicketUser,
        related_name = 'managed_ticketproject',
        verbose_name = _("Managers")
    )

    staff = models.ManyToManyField (
        TicketUser,
        related_name = 'assigned_ticketprojects',
        verbose_name = _("Staff")
    )

    ## Campi
    assistance_email = models.EmailField(
        verbose_name = _("Assistance Email"),
        blank = True,
        null = True,
    )
    icon = models.ImageField(
        blank = True, null = True,
        upload_to = settings.TICKETPROJECT_PATH_ICON,
        verbose_name = _("Icon")
    )
    default_for_organization = models.BooleanField(
        verbose_name = _("Default for organization"),
        help_text = _("This will be the default project when a guest insert a ticket"),
        default = False,
    )

    objects = ProjectManager()
    all_objects = AllProjectManager()

    def get_string_managers(self):
        manager = [ m.user.username for m in self.managers.all() ]
        return ', '.join(manager)

    def get_email_managers(self):
        manager_emails = [ m.user.email for m in self.managers.all() if m.user.email ]
        return manager_emails

    def get_list_managers(self):
        return self.managers.all()

    def get_string_staff(self):
        staff = [ m.user.username for m in self.staff.all() ]
        return ', '.join(staff)

    def get_list_staff(self):
        return self.staff.all()

    def get_full_name(self):
        return self.tickets.count()
    full_name=property(get_full_name)

    def get_total_tickets(self):
        return self.tickets.count()
    total_tickets=property(get_total_tickets)
    get_total_tickets.short_description = _('Total Tickets')

    def get_open_tickets(self):
        return self.tickets.filter(status=7).count()
    open_tickets=property(get_open_tickets)
    get_open_tickets.short_description = _('Open Tickets')

    def get_close_tickets(self):
        return self.total_tickets - self.open_tickets
    close_tickets=property(get_close_tickets)
    get_close_tickets.short_description = _('Close Tickets')

    def get_percentage_done(self):
        percentage = 100
        if self.total_tickets:
            percentage = (100 * self.close_tickets) / self.total_tickets
        return int(percentage)
    percentage_done=property(get_percentage_done)
    get_percentage_done.short_description = _('Percentage Done')

    def get_total_work_hours(self):
        """
        restituisce le ore totali
        """
        assistances =  TicketAssistance.objects.filter(ticket__project=self)
        totale_ore = timedelta(0)
        for assistance in assistances:
            totale_ore += assistance.real_duration
        s = totale_ore.days * 86400 + totale_ore.seconds
        hours, remainder = divmod(s, 3600)
        minutes, seconds = divmod(remainder, 60)
        str_hours = 'ora'
        if hours > 1:
            str_hours = 'ore'
        str_minutes = 'minuto'
        if minutes > 1:
            str_minutes = 'minuti'
        return '%s %s, %s %s' % (hours, str_hours, minutes, str_minutes)
    total_work_hours=property(get_total_work_hours)
    get_total_work_hours.short_description = _("Total Work Hours")

    def get_total_assistances(self):
        return TicketAssistance.objects.filter(ticket__project=self).count()
    total_assistances=property(get_total_assistances)
    get_total_assistances.short_description = _("Total Assistances")

    def get_assistances(self):
        """
        restituisce tutte le assistenze fatte sul progetto
        """
        return TicketAssistance.objects.filter(ticket__project=self)


    def get_assistances_by_worker(self):
        """
        restituisce le ore per ogni lavoratore
        """

        worker = None
        workers = {}
        cost_per_hour = 0
        extra_per_hour = 0
        assistances = self.get_assistances().order_by('worker')
        for ass in assistances:
            if worker == ass.worker:
                workers[ass.worker]['work_duration'] += (ass.get_durations()[0])/60.00
                workers[ass.worker]['extra_duration'] += (ass.get_durations()[1])/60.00
                workers[worker]['work_cost'] = cost_per_hour * workers[worker]['work_duration']
                workers[worker]['extra_cost'] = extra_per_hour * workers[worker]['extra_duration']
                workers[worker]['total_cost'] = workers[worker]['work_cost'] + workers[worker]['extra_cost']
            if worker != ass.worker:
                worker = ass.worker
                # se sono impostati i costi del personale calcolo anche il costo della MDO
                if "timereport" in django_settings.INSTALLED_APPS:
                    employee = worker.user.employee.all()
                    if employee:
                        employee = employee[0]
                        # al primo giro 'worker' e' uguale a None
                        cost_per_hour = employee.cost or 0
                        extra_per_hour = employee.extra_cost or 0

                workers[ass.worker] = {
                    'user': ass.worker.user,
                    'work_duration': 0,
                    'extra_duration': 0,
                    'work_cost': 0,
                    'extra_cost': 0,
                    'total_cost': 0
                }
                workers[ass.worker]['work_duration'] = ass.get_durations()[0]/60.00
                workers[ass.worker]['extra_duration'] = ass.get_durations()[1]/60.00
                workers[worker]['work_cost'] = cost_per_hour * workers[worker]['work_duration']
                workers[worker]['extra_cost'] = extra_per_hour * workers[worker]['extra_duration']
                workers[worker]['total_cost'] = workers[worker]['work_cost'] + workers[worker]['extra_cost']

        return workers

    def get_totals_assistances(self):
        """
        restituisce i totali MDO
        """
        totals={
            'work_duration': 0,
            'extra_duration': 0,
            'work_cost': 0,
            'extra_cost': 0,
            'total_cost': 0
        }
        for worker in self.get_assistances_by_worker().values():
            totals['work_duration'] += worker['work_duration']
            totals['extra_duration'] += worker['extra_duration']
            totals['work_cost'] += worker['work_cost']
            totals['extra_cost'] += worker['extra_cost']
            totals['total_cost'] += worker['total_cost']
        return totals

    class Meta:
        verbose_name = _("Project")
        verbose_name_plural = _("Projects")
        permissions = (
            ("view_ticketproject"         , "Can view ticket projects"),
            ("list_all_ticket_projects"   , "Can list all ticket projects"),
        )
        ordering = ['name']

    def __unicode__(self):
        if self.organization:
            return "%s (%s)" % (unicode(self.name), unicode(self.organization))
        else:
            return unicode(self.name)


###########################
# CATEGORY                #
###########################
CSTATUS_CODES = (
    (0, _('Disable')),
    (1, _('Enable')),
)

class Category(DateModel, UserModel, StatusModel):
    ## Campi
    name = models.CharField (
        max_length=30,
        verbose_name=_("Name"),
    )
    description = models.TextField (
        null=True,
        verbose_name=_("Description"),
    )
    def __unicode__(self):
        return self.name

    class Meta:
        verbose_name = _("Category")
        verbose_name_plural = _("Categories")

class CategoryTranslation(TranslationModel):
    category = models.ForeignKey(
        Category,
        related_name='translations',
    )
    name = models.CharField (
        max_length=30,
        verbose_name=_("Name"),
    )

    class Meta:
        verbose_name = _('Translation')
        verbose_name_plural = _('Translations')
        unique_together = (('category', 'language',),)

    def __unicode__(self):
        return self.name


###########################
# CATEGORY                #
###########################
class Tipology(DateModel, UserModel, StatusModel):
    ## Campi
    name = models.CharField (
        max_length=30,
        verbose_name=_("Name"),
    )
    description = models.TextField (
        null=True,
        verbose_name=_("Description"),
    )

    def __unicode__(self):
        return self.name

    class Meta:
        verbose_name = _("Tipology")
        verbose_name_plural = _("Tipologies")

class TipologyTranslation(TranslationModel):
    tipology = models.ForeignKey(
        Tipology,
        related_name='translations',
    )
    name = models.CharField (
        max_length=30,
        verbose_name=_("Name"),
    )

    class Meta:
        verbose_name = _('Translation')
        verbose_name_plural = _('Translations')
        unique_together = (('tipology', 'language',),)

    def __unicode__(self):
        return self.name


#########################################
# RIGA ASSISTENZA                       #
#########################################
TICKETASSISTANCE_CARRYING = (
    ('com', _('company')),
    ('own', _('own')),
    ('cus', _('customer')),
    ('pub', _('public')),
    ('oth', _('other')),
)

class TicketAssistanceManager(models.Manager):
    def get_query_set(self):
        qs = super(TicketAssistanceManager, self).get_query_set()
        user = get_current_user()

        if not user:
            return qs.none()

        if not user.is_authenticated():
            return qs.none()

        if user.is_superuser or user.has_perm('ticket.list_all_ticketassistances'):
            return qs

        qs = qs.filter(worker__user=user)
        return qs

class TicketTicketAssistanceManager(models.Manager):
    """
        Assistenze relative ai ticket. Torna tutte le assistenze dei propri ticket,
        anche quelle degli altri utenti.
    """
    def get_query_set(self):
        qs = super(TicketTicketAssistanceManager, self).get_query_set()
        user = get_current_user()

        if not user:
            return qs.none()

        if not user.is_authenticated():
            return qs.none()

        if user.is_superuser or user.has_perm('ticket.list_all_ticketassistances'):
            return qs

        t = Ticket.objects.all()
        qs = qs.filter(ticket__in=t)
        return qs

class AllTicketAssistanceManager(models.Manager):
    """
    Restituisce tutte le assistenze
    """
    def get_query_set(self):
        return super(AllTicketAssistanceManager, self).get_query_set()


class TicketAssistance (DateModel, UserModel, StatusModel):
    ## Foreignkey
    ticket = models.ForeignKey (
        Ticket,
        related_name = 'assistances',
        db_index = True,
        verbose_name = _("ticket"),
    )
    worker = models.ForeignKey (
        TicketUser,
        related_name = 'assistances',
        db_index = True,
        verbose_name = _("worker"),
    )
    ## Campi Modello
    date_begin = models.DateTimeField (
        db_index = True,
        blank = True, null = True,
        verbose_name = _("date begin"),
    )
    date_end = models.DateTimeField (
        db_index = True,
        blank = True, null = True,
        verbose_name = _("date end"),
    )
    work_duration = models.IntegerField (
        blank = True, null = True,
        verbose_name = _("work duration"),
        help_text = _("In Minutes")
    )
    job_description = jumbo_models.HTMLField (
        blank = True, null = True,
        verbose_name = _("Job Description"),
    )
    extra_duration = models.IntegerField (
        verbose_name = _("extra duration"),
        help_text = _("In Minutes"),
        blank = True, null = True,
    )
    travel_duration = models.IntegerField (
        verbose_name = _("travel duration"),
        help_text = _("In Minutes"),
        blank = True, null = True,
    )
    kilometers = models.IntegerField (
        verbose_name = _("kilometers"),
        blank = True, null = True,
    )
    overnight_stay = models.BooleanField (
        verbose_name = _("overnight stay"),
    )
    carrying = models.CharField (
        max_length = 3,
        choices = TICKETASSISTANCE_CARRYING,
        db_index = True,
        verbose_name = _("carrying"),
        blank = True, null = True,
    )
    pay_exit = models.BooleanField (
        db_index = True,
        verbose_name = _("pay exit"),
    )
    procedure = models.BooleanField (
        ## preventivo già trasformato in procedure
        default = False,
        db_index = True,
        verbose_name = _("Ordered"),
    )
    objects = TicketAssistanceManager()
    ticket_assistances = TicketTicketAssistanceManager()
    all_assistances = AllTicketAssistanceManager()

    def save(self, force_insert=False, force_update=False):
        if not self.work_duration and self.date_end and self.date_begin:
            ## calcolo la durata in minuti
            work_duration = self.get_durations()[0]
            time_round = get_settings("ticket.settings", "TICKETASSISTANCE_TIME_ROUND")
            time_step =  get_settings("ticket.settings", "TICKETASSISTANCE_TIME_STEP")
            if (
                isinstance(time_round, (int,)) and isinstance(time_step, (int,))
                and time_round > 0 and time_step > 0
            ):
                waste_time = work_duration % 60
                first_step = waste_time / time_round * time_round
                round = waste_time % time_round / time_step
                self.work_duration = first_step + round * time_round + work_duration / 60 * 60

        if not self.extra_duration and self.date_end and self.date_begin:
            extra_duration = self.get_durations()[1]
            time_round = get_settings("ticket.settings", "TICKETASSISTANCE_TIME_ROUND")
            time_step =  get_settings("ticket.settings", "TICKETASSISTANCE_TIME_STEP")
            if (
                isinstance(time_round, (int,)) and isinstance(time_step, (int,))
                and time_round > 0 and time_step > 0
            ):
                waste_time = extra_duration % 60
                first_step = waste_time / time_round * time_round
                round = waste_time % time_round / time_step
                self.extra_duration = first_step + round * time_round + extra_duration / 60 * 60

        ## verifico se è stato impostata la costante LIMIT_INSERT_ASSISTANCE
        ## valore espresso in giorni oltre il quale l'assistenza non è più inseribile o modificabile
        limit_insert_assistance = get_settings("ticket.settings", "LIMIT_INSERT_ASSISTANCE")
        if limit_insert_assistance:
            limit_insert_date = (self.ticket.deadline or self.ticket.date_create) + timedelta(days=limit_insert_assistance)
            user = get_current_user()
            if limit_insert_date.date() < date.today() and not user.is_superuser:
                raise Exception("Limit insert is exceeded expiration date. Contact Administrator")

        super(TicketAssistance, self).save(force_insert=force_insert, force_update=force_update)

    def get_real_duration(self):
        real_duration = timedelta(0)
        if self.date_begin and self.date_end:
            real_duration = self.date_end - self.date_begin
        return real_duration
    real_duration=property(get_real_duration)


    def get_calculated_duration(self):
        ## espressi in minuti
        work_duration = self.work_duration or 0
        extra_duration = self.extra_duration or 0
        travel_duration = self.travel_duration or 0
        calc_duration = timedelta(seconds=(work_duration + extra_duration + travel_duration)*60)
        return calc_duration
    calculated_duration=property(get_calculated_duration)

    def get_durations(self):
        """
        restituisce il tempo ordinario e straordinario di lavoro effettuato
        il valore è espresso in minuti
        """
        from datetime import time, datetime, timedelta
        time_begin = time(self.date_begin.hour,self.date_begin.minute)
        time_end = time(self.date_end.hour,self.date_end.minute)
        work_duration = timedelta(hours=0,minutes=0)
        duration = timedelta(hours=time_end.hour,minutes=time_end.minute) - timedelta(hours=time_begin.hour,minutes=time_begin.minute)

        weekday = datetime.today().weekday()

        ## se esiste l'applicazione timereport prendo gli orari di lavoro del
        ## lavoratore - worker dal proprio profilo
        if "timereport" in django_settings.INSTALLED_APPS:
            working_time = []
            from timereport.models import TimeProfile
            profiles = TimeProfile.objects.filter(employee__user = self.worker.user, weekday=weekday)
            if profiles:
                profiles=profiles.values('time_in', 'time_out')
                for profile in profiles:
                    working_time.append(profile['time_in'])
                    working_time.append(profile['time_out'])

        ## se non esiste l'applicazione timereport oppure il lavoratore non ha un profilo
        ## prendo i valori di default impostati nella costante
        if not "timereport" in django_settings.INSTALLED_APPS or not working_time:
            TICKETASSISTANCE_WORKING_TIME = get_settings("ticket.settings", "TICKETASSISTANCE_WORKING_TIME")
            working_time = TICKETASSISTANCE_WORKING_TIME[weekday]

        for num, val_time in enumerate(working_time):
            time_mod = timedelta(hours=val_time.hour,minutes=val_time.minute)

            if num%2 == 0:
                ## ora inizio
                time_begin_mod = timedelta(hours=time_begin.hour,minutes=time_begin.minute)
                if time_begin_mod < time_mod:
                    time_begin_mod = time_mod
            else:
                ## ora fine
                time_end_mod = timedelta(hours=time_end.hour,minutes=time_end.minute)
                if time_end_mod > time_mod:
                    time_end_mod = time_mod
                if time_end_mod <= time_begin_mod:
                    continue
                work_duration = work_duration + time_end_mod - time_begin_mod

        extra_duration = duration - work_duration
        return work_duration.seconds/60, extra_duration.seconds/60

    def get_status(self):
        status = ugt("open")
        if self.date_begin and self.date_end:
            status = ugt("close")
        return status
    get_status=property(get_status)


    def can_be_delete(self):
        from jumbo.middleware.thread_local import get_current_user
        user = get_current_user()

        if user.has_perm("ticket.delete_all_ticketassistances"):
            return True

        if user.has_perm("ticket.delete_ticketassistance") and self.worker.user == user:
            return True

        return False

    def delete(self, *args, **kws):
        if self.can_be_delete():
            return super(TicketAssistance, self).delete(*args, **kws)
        raise PermissionDenied


    if "procedure" in django_settings.INSTALLED_APPS:
        @transaction.commit_manually
        def create_or_append_procedure(self, procedure = None, date_p = None):
            """
            crea o aggiunge ad una Pratica le Righe Preventivo
            """
            if not date_p:
                date_p = date.today()
            process_open_assistance = get_settings("ticket.settings", "TICKETASSISTANCE_PROCESS_OPEN_ASSISTANCE")
            ## se il ticket non è chiuso esco
            if self.ticket.status != 0 and not process_open_assistance:
                transaction.rollback()
                return (None, _("Ticket %s is still open") % self.ticket.id)
            ## se l'assistenza è già stata inserita in un ordine esco
            if self.procedure:
                transaction.rollback()
                return (None, _("Assistance %s already ordered") % self.ticket.id)

            ## se l'assistenza è ancora aperta esco
            if not self.get_status == "close":
                transaction.rollback()
                return (None, _("Assistance %s is still open") % self.ticket.id)

            ## se non esiste prodotto assistenza esco
            ## occorre settare TICKETASSISTANCE_PRODUCT_CODE per associare TicketAssistance
            ## a Product
            try:
                from product.models import Product
                code = get_settings("ticket.settings", "TICKETASSISTANCE_PRODUCT_CODE")
                product = Product.objects.get(code=code)
            except:
                transaction.rollback()
                return (None, "TICKETASSISTANCE_PRODUCT_CODE is not configured")

            vat = self.ticket.organization.vat or product.vat
            ## se non esiste creo l'ordine
            if not procedure:
                from procedure.models import Procedure
                description = get_settings("ticket.settings", "TICKETASSISTANCE_DESCRIPTION")
                try:
                    storecause = product.store_cause
                except:
                    transaction.rollback()
                    return (None, _("Invalid store cause for product %s") % product.id)
                procedure = Procedure(
                    owner=self.ticket.organization, recipient=self.ticket.organization,
                    payment_form=self.ticket.organization.payment_form,
                    office=self.ticket.organization.office, store_cause=storecause,
                    contact=self.ticket.contact, description=description, date_p=date_p
                )
                procedure.save()

            ## creo la riga ordine
            from procedure.models import ProcedureRow
            description = ugt('Ticket n. %s of %s' % (self.ticket.id, self.date_begin.strftime('%d/%m/%Y')))
            ## tempi
            work_duration = self.work_duration or 0
            extra_duration = self.extra_duration or 0
            travel_duration = self.travel_duration or 0

            try:
                project = self.ticket.project.project_ptr
            except:
                project = None

            if work_duration >= 0:
                procedurerow1 = ProcedureRow(
                    status=1, procedure=procedure, product=product, vat=vat,
                    store_cause=product.store_cause, quantity=decimal.Decimal(work_duration)/60,
                    m_unit = product.m_unit, project = project,
                    description='Assistance #%s for %s' % (self.id, description), long_description=self.job_description,
                )
                procedurerow1.save()
            if travel_duration > 0:
                procedurerow2 = ProcedureRow(
                    status=1, procedure=procedure, product=product, vat=vat,
                    store_cause=product.store_cause, quantity=decimal.Decimal(travel_duration)/60,
                    m_unit = product.m_unit, project = project,
                    description='Travel for #%s for %s' % (self.id, description), long_description=self.job_description,
                )
                procedurerow2.save()
            if extra_duration > 0:
                procedurerow3 = ProcedureRow(
                    status=1, procedure=procedure, product=product, vat=vat,
                    store_cause=product.store_cause, quantity=decimal.Decimal(extra_duration)/60,
                    m_unit = product.m_unit, project = project,
                    description='Extra for #%s for %s' % (self.id, description), long_description=self.job_description,
                )
                procedurerow3.save()
            ## costi
            #
            if "timereport" in django_settings.INSTALLED_APPS:
                employee = self.worker.user.employee.all()
                if employee:
                    employee = employee[0]
                    cost_per_hour = employee.cost or 0
                    extra_cost_per_hour = employee.extra_cost or 0
                    if work_duration >= 0:
                        procedurerow1.price = cost_per_hour
                        procedurerow1.save()

                    if travel_duration > 0:
                        procedurerow2.price = cost_per_hour
                        procedurerow2.save()

                    if extra_duration > 0:
                        procedurerow3.price = extra_cost_per_hour
                        procedurerow3.save()

            if procedure.rows.count() == 0:
                transaction.rollback()
                return (None, "Error!!!")

            ## Riga Assistenza trasformata correttamente
            self.procedure = True
            self.save()
            transaction.commit()
            return (procedure, "")

    class Meta:
        verbose_name = _("assistance ticket")
        verbose_name_plural = _("assistances ticket")
        ordering = ('-date_begin',)
        permissions = (
            ("view_ticketassistance",          "Can view ticket assistances"),
            ("list_all_ticketassistances",     "Can list all ticket assistances"),
            ("modify_all_ticketassistances",   "Can modify all ticket assistances"),
            ("delete_all_ticketassistances",   "Can delete all ticket assistances"),
        )

    def __unicode__(self):
        return u"%s --> %s" % (self.ticket, self.worker)


class Log(DateModel, UserModel, StatusModel):
    ticket = models.ForeignKey (
        Ticket,
        related_name = 'logs',
        db_index = True,
        verbose_name = _("Ticket"),
    )
    field = models.CharField(
        max_length=100,
        verbose_name=_("Field"),
    )
    old_value = models.TextField(
        verbose_name=_("Old Value"),
    )
    new_value = models.TextField(
        verbose_name=_("New Value"),
    )

    class Meta:
        verbose_name = _("ticket log")
        verbose_name_plural = _("ticket logs")

    def __unicode__(self):
        return u"%s --> modificato %s" % (self.ticket, self.field)


class Settings(JSettings):
    pass
